@@ -1,6 +1,6 @@
 /* wl-clipboard
  *
- * Copyright © 2019 Sergey Bugaev <bugaevc@gmail.com>
+ * Copyright © 2018-2023 Sergey Bugaev <bugaevc@gmail.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,4 +27,13 @@
 #    include "xdg-shell.h"
 #endif
 
+#ifdef HAVE_GTK_SHELL
+#    include "gtk-shell.h"
+#endif
+
+/* Not strictly speaking a shell */
+#ifdef HAVE_XDG_ACTIVATION
+#    include "xdg-activation.h"
+#endif
+
 #endif /* INCLUDES_SHELL_PROTOCOLS_H */
