@@ -29,12 +29,14 @@
 
 #define BIND(interface_name, known_version) \
 if (strcmp(interface, #interface_name) == 0) { \
-    self->interface_name = wl_registry_bind( \
-        wl_registry, \
-        name, \
-        &interface_name ## _interface, \
-        known_version \
-    ); \
+    if (version >= (known_version)) { \
+        self->interface_name = wl_registry_bind( \
+            wl_registry, \
+            name, \
+            &interface_name ## _interface, \
+            known_version \
+        ); \
+    } \
 }
 
 static void wl_registry_global_handler(
@@ -81,7 +83,7 @@ static void wl_registry_global_handler(
     BIND(zwlr_data_control_manager_v1, version > 2 ? 2 : version)
 #endif
 
-    if (strcmp(interface, "wl_seat") == 0) {
+    if (strcmp(interface, "wl_seat") == 0 && version >= 2) {
         struct seat *seat = calloc(1, sizeof(struct seat));
         seat->proxy = wl_registry_bind(
             wl_registry,
