@@ -1,6 +1,6 @@
 /* wl-clipboard
  *
- * Copyright © 2019 Sergey Bugaev <bugaevc@gmail.com>
+ * Copyright © 2018-2023 Sergey Bugaev <bugaevc@gmail.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +19,9 @@
 #ifndef UTIL_FILES_H
 #define UTIL_FILES_H
 
+struct wl_display;
+void complain_about_closed_stdio(struct wl_display *wl_display);
+
 int create_anonymous_file(void);
 
 void trim_trailing_newline(const char *file_path);
