@@ -1,6 +1,6 @@
 /* wl-clipboard
  *
- * Copyright © 2019 Sergey Bugaev <bugaevc@gmail.com>
+ * Copyright © 2018-2023 Sergey Bugaev <bugaevc@gmail.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 #include "types/device-manager.h"
 #include "types/registry.h"
 #include "types/popup-surface.h"
+#include "types/offer.h"
 
 #include "util/files.h"
 #include "util/string.h"
@@ -30,8 +31,10 @@
 #include <wayland-client.h>
 #include <unistd.h>
 #include <string.h>
+#include <fcntl.h> // open
 #include <libgen.h>
 #include <getopt.h>
+#include <signal.h>
 
 static struct {
     int stay_in_foreground;
@@ -52,8 +55,29 @@ static void did_set_selection_callback(struct copy_action *copy_action) {
         /* Move to background.
          * We fork our process and leave the
          * child running in the background,
-         * while exiting in the parent.
+         * while exiting in the parent. We also
+         * replace stdin/stdout with /dev/null
+         * so the stdout file descriptor isn't
+         * kept alive, and chdir to the root, to
+         * prevent blocking file systems from
+         * being unmounted.
          */
+        int devnull = open("/dev/null", O_RDWR);
+        if (devnull >= 0) {
+            dup2(devnull, STDOUT_FILENO);
+            dup2(devnull, STDIN_FILENO);
+            close(devnull);
+        } else {
+            /* If we cannot open /dev/null,
+             * just close stdin/stdout.
+             */
+            close(STDIN_FILENO);
+            close(STDOUT_FILENO);
+        }
+        if (chdir("/") < 0) {
+            perror("chdir /");
+        }
+        signal(SIGHUP, SIG_IGN);
         pid_t pid = fork();
         if (pid < 0) {
             perror("fork");
@@ -65,28 +89,20 @@ static void did_set_selection_callback(struct copy_action *copy_action) {
     }
 }
 
-static void cleanup_and_exit(struct copy_action *copy_action, int code) {
-    /* We're done copying!
-     * All that's left to do now is to
-     * clean up after ourselves and exit.*/
-    char *temp_file = (char *) copy_action->file_to_copy;
-    if (temp_file != NULL) {
-        /* Clean up our temporary file */
-        execlp("rm", "rm", "-r", dirname(temp_file), NULL);
-        perror("exec rm");
-        exit(1);
-    } else {
-        exit(code);
-    }
-}
-
 static void cancelled_callback(struct copy_action *copy_action) {
-    cleanup_and_exit(copy_action, 0);
+    exit(0);
 }
 
 static void pasted_callback(struct copy_action *copy_action) {
     if (options.paste_once) {
-        cleanup_and_exit(copy_action, 0);
+        exit(0);
+    }
+}
+
+static void selection_callback(struct offer *offer, int primary) {
+    /* We're not interested */
+    if (offer != NULL) {
+        offer_destroy(offer);
     }
 }
 
@@ -100,7 +116,7 @@ static void print_usage(FILE *f, const char *argv0) {
         "Options:\n"
         "\t-o, --paste-once\tOnly serve one paste request and then exit.\n"
         "\t-f, --foreground\tStay in the foreground instead of forking.\n"
-        "\t-c, --clear\t\tInstead of copying anything, clear the clipboard.\n"
+        "\t-c, --clear\t\tInstead of copying, clear the clipboard.\n"
         "\t-p, --primary\t\tUse the \"primary\" clipboard.\n"
         "\t-n, --trim-newline\tDo not copy the trailing newline character.\n"
         "\t-t, --type mime/type\t"
@@ -183,9 +199,20 @@ static void parse_options(int argc, argv_t argv) {
 int main(int argc, argv_t argv) {
     parse_options(argc, argv);
 
+    /* Ignore SIGPIPE.
+     * We don't really output anything
+     * to our stdout, yet we don't want
+     * to get killed when writing clipboard
+     * contents to a closed pipe.
+     */
+    signal(SIGPIPE, SIG_IGN);
+
     struct wl_display *wl_display = wl_display_connect(NULL);
     if (wl_display == NULL) {
-        bail("Failed to connect to a Wayland server");
+        complain_about_wayland_connection();
+    }
+    if (wl_display_get_fd(wl_display) <= STDERR_FILENO) {
+        complain_about_closed_stdio(wl_display);
     }
 
     struct registry *registry = calloc(1, sizeof(struct registry));
@@ -197,11 +224,7 @@ int main(int argc, argv_t argv) {
 
     struct seat *seat = registry_find_seat(registry, options.seat_name);
     if (seat == NULL) {
-        if (options.seat_name != NULL) {
-            bail("No such seat");
-        } else {
-            bail("Missing a seat");
-        }
+        complain_about_missing_seat(options.seat_name);
     }
 
     /* Create the device */
@@ -212,6 +235,7 @@ int main(int argc, argv_t argv) {
     }
 
     struct device *device = device_manager_get_device(device_manager, seat);
+    device->selection_callback = selection_callback;
 
     if (!device_supports_selection(device, options.primary)) {
         complain_about_selection_support(options.primary);
@@ -219,6 +243,7 @@ int main(int argc, argv_t argv) {
 
     /* Create and initialize the copy action */
     struct copy_action *copy_action = calloc(1, sizeof(struct copy_action));
+    copy_action->fd_to_copy_from = -1;
     copy_action->device = device;
     copy_action->primary = options.primary;
 
@@ -241,7 +266,28 @@ int main(int argc, argv_t argv) {
             if (options.mime_type == NULL) {
                 options.mime_type = infer_mime_type_from_contents(temp_file);
             }
-            copy_action->file_to_copy = temp_file;
+            copy_action->fd_to_copy_from = open(
+                temp_file,
+                O_RDONLY | O_CLOEXEC
+            );
+            if (copy_action->fd_to_copy_from < 0) {
+                perror("Failed to open temp file");
+                return 1;
+            }
+            /* Now, remove the temp file and its
+             * containing directory. We still keep
+             * access to the file through our open
+             * file descriptor.
+             */
+            int rc = unlink(temp_file);
+            if (rc < 0) {
+                perror("Failed to unlink temp file");
+            }
+            rc = rmdir(dirname(temp_file));
+            if (rc < 0) {
+                perror("Failed to remove temp file directory");
+            }
+            free(temp_file);
         }
 
         /* Create the source */
@@ -275,6 +321,5 @@ int main(int argc, argv_t argv) {
     while (wl_display_dispatch(wl_display) >= 0);
 
     perror("wl_display_dispatch");
-    cleanup_and_exit(copy_action, 1);
     return 1;
 }
