@@ -1,6 +1,6 @@
 /* wl-clipboard
  *
- * Copyright © 2019 Sergey Bugaev <bugaevc@gmail.com>
+ * Copyright © 2018-2023 Sergey Bugaev <bugaevc@gmail.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -28,6 +28,8 @@
 
 #include <wayland-client.h>
 #include <unistd.h>
+#include <assert.h>
+#include <fcntl.h>
 #include <errno.h>
 #include <string.h>
 #include <getopt.h>
@@ -58,6 +60,7 @@ struct types {
 
 static struct wl_display *wl_display = NULL;
 static struct popup_surface *popup_surface = NULL;
+static int offer_received = 0;
 
 static struct types classify_offer_types(struct offer *offer) {
     struct types types = { 0 };
@@ -171,7 +174,71 @@ static const char *mime_type_to_request(struct types types) {
 #undef try_any_text
 #undef try_any
 
+static int run_paste_command(int stdin_fd, const char *clipboard_state) {
+    /* Spawn a cat to perform the copy.
+     * If watch mode is active, we spawn
+     * a custom command instead.
+     */
+    pid_t pid = fork();
+    if (pid < 0) {
+        perror("fork");
+        close(stdin_fd);
+        return 0;
+    }
+    if (pid == 0) {
+        dup2(stdin_fd, STDIN_FILENO);
+        close(stdin_fd);
+        if (options.watch) {
+            if (clipboard_state != NULL) {
+                setenv("CLIPBOARD_STATE", clipboard_state, 1);
+            }
+            execvp(options.watch_command[0], options.watch_command);
+            fprintf(
+                stderr,
+                "Failed to spawn %s: %s",
+                options.watch_command[0],
+                strerror(errno)
+            );
+        } else {
+            execlp("cat", "cat", NULL);
+            perror("exec cat");
+        }
+        exit(1);
+    }
+    close(stdin_fd);
+    waitpid(pid, NULL, 0);
+    return 1;
+}
+
+static void complain_no_suitable_type() {
+    fprintf(stderr, "Clipboard content is not available as ");
+    if (options.explicit_type != NULL) {
+        fprintf(stderr, "requested type \"%s\"\n", options.explicit_type);
+    } else {
+        assert(options.inferred_type);
+        fprintf(
+            stderr,
+            "inferred output type \"%s\"\n",
+            options.inferred_type
+        );
+    }
+    fprintf(stderr, "Use \"wl-paste --list-types\" to view available types.");
+    if (options.explicit_type == NULL) {
+        fprintf(stderr, " Use \"--type\" to explicitly specify a type.");
+    }
+    fputc('\n', stderr);
+    exit(1);
+}
+
 static void selection_callback(struct offer *offer, int primary) {
+    /* Ignore all but the first non-NULL offer.
+     * This could happen due to reentrancy, though
+     * we try to prevent it in other ways.
+     */
+    if (offer_received && !options.watch) {
+        return;
+    }
+
     /* Ignore events we're not interested in */
     if (primary != options.primary) {
         if (offer != NULL) {
@@ -181,12 +248,20 @@ static void selection_callback(struct offer *offer, int primary) {
     }
 
     if (offer == NULL) {
-        if (options.watch) {
+        if (!options.watch) {
+            bail("Nothing is copied");
+        }
+        int devnull = open("/dev/null", O_RDONLY | O_CLOEXEC);
+        if (devnull < 0) {
+            perror("open /dev/null");
             return;
         }
-        bail("No selection");
+        run_paste_command(devnull, "nil");
+        return;
     }
 
+    offer_received = 1;
+
     if (options.list_types) {
         offer_for_each_mime_type(offer, mime_type) {
             printf("%s\n", mime_type);
@@ -202,7 +277,7 @@ static void selection_callback(struct offer *offer, int primary) {
             offer_destroy(offer);
             return;
         }
-        bail("No suitable type of content copied");
+        complain_no_suitable_type();
     }
 
     /* Never append a newline character to binary content */
@@ -230,45 +305,24 @@ static void selection_callback(struct offer *offer, int primary) {
         popup_surface_destroy(popup_surface);
         popup_surface = NULL;
     }
-    wl_display_roundtrip(wl_display);
-
-    /* Spawn a cat to perform the copy.
-     * If watch mode is active, we spawn
-     * a custom command instead.
+    /* Make sure the receive request reaches
+     * the compositor before we block on reading.
+     * We call flush() instead of dispatch() to
+     * prevent reentrancy.
      */
-    pid_t pid = fork();
-    if (pid < 0) {
-        perror("fork");
+    wl_display_flush(wl_display);
+
+    close(pipefd[1]);
+    rc = run_paste_command(pipefd[0], "data");
+    if (!rc) {
         if (options.watch) {
             /* Try to cope without exiting completely */
-            close(pipefd[0]);
-            close(pipefd[1]);
             offer_destroy(offer);
             return;
         }
         exit(1);
     }
-    if (pid == 0) {
-        dup2(pipefd[0], STDIN_FILENO);
-        close(pipefd[0]);
-        close(pipefd[1]);
-        if (options.watch) {
-            execvp(options.watch_command[0], options.watch_command);
-            fprintf(
-                stderr,
-                "Failed to spawn %s: %s",
-                options.watch_command[0],
-                strerror(errno)
-            );
-        } else {
-            execlp("cat", "cat", NULL);
-            perror("exec cat");
-        }
-        exit(1);
-    }
-    close(pipefd[0]);
-    close(pipefd[1]);
-    wait(NULL);
+
     if (!options.no_newline && !options.watch) {
         rc = write(STDOUT_FILENO, "\n", 1);
         if (rc != 1) {
@@ -418,7 +472,10 @@ int main(int argc, argv_t argv) {
 
     wl_display = wl_display_connect(NULL);
     if (wl_display == NULL) {
-        bail("Failed to connect to a Wayland server");
+        complain_about_wayland_connection();
+    }
+    if (wl_display_get_fd(wl_display) <= STDERR_FILENO) {
+        complain_about_closed_stdio(wl_display);
     }
 
     struct registry *registry = calloc(1, sizeof(struct registry));
@@ -430,11 +487,7 @@ int main(int argc, argv_t argv) {
 
     struct seat *seat = registry_find_seat(registry, options.seat_name);
     if (seat == NULL) {
-        if (options.seat_name != NULL) {
-            bail("No such seat");
-        } else {
-            bail("Missing a seat");
-        }
+        complain_about_missing_seat(options.seat_name);
     }
 
     /* Create the device */
